# usage_stats.py

from typing import Dict

class UsageStats:
    def __init__(self, model=None):
        self.model = model
        self.total_usage = {
            'input_tokens': 0,
            'output_tokens': 0,
            'total_tokens': 0,
            'input_cost': 0.0,
            'output_cost': 0.0,
            'total_cost': 0.0
        }
        self.operation_usage: Dict[str, Dict[str, float]] = {}

    def update(self, meta, operation_name):
        # Update total usage
        self.total_usage['input_tokens'] += meta.get('input_tokens', 0)
        self.total_usage['output_tokens'] += meta.get('output_tokens', 0)
        self.total_usage['total_tokens'] += meta.get('total_tokens', 0)
        self.total_usage['input_cost'] += meta.get('input_cost', 0.0)
        self.total_usage['output_cost'] += meta.get('output_cost', 0.0)
        self.total_usage['total_cost'] += meta.get('total_cost', 0.0)
        self.total_usage['total_cost'] = round(self.total_usage['total_cost'], 5)

        # Update per-operation usage
        if operation_name not in self.operation_usage:
            self.operation_usage[operation_name] = {
                'input_tokens': 0,
                'output_tokens': 0,
                'total_tokens': 0,
                'input_cost': 0.0,
                'output_cost': 0.0,
                'total_cost': 0.0
            }

        op_usage = self.operation_usage[operation_name]
        op_usage['input_tokens'] += meta.get('input_tokens', 0)
        op_usage['output_tokens'] += meta.get('output_tokens', 0)
        op_usage['total_tokens'] += meta.get('total_tokens', 0)
        op_usage['input_cost'] += meta.get('input_cost', 0.0)
        op_usage['output_cost'] += meta.get('output_cost', 0.0)
        op_usage['total_cost'] += meta.get('total_cost', 0.0)
        op_usage['total_cost'] = round(op_usage['total_cost'], 5)

    def to_dict(self):
        return {
            'model': self.model,
            'total_usage': self.total_usage,
            'operation_usage': self.operation_usage
        }



# # postprocessor.py
#
# import logging
# from typing import Any, Optional, Dict, List, Type
#
# from .schemas import PostprocessingResult, GenerationResult  # Ensure these are correctly imported
# from string2dict import String2Dict  # Import String2Dict
# import json
#
# # Base class for processing steps
# class ProcessingStep:
#     def __init__(self, params: Optional[Dict[str, Any]] = None):
#         self.params = params or {}
#
#     def process(self, data: Any) -> PostprocessingResult:
#         raise NotImplementedError("Each step must implement the process method.")
#
# # Define specific processing steps
# class SemanticIsolationStep(ProcessingStep):
#     def process(self, data: Any) -> PostprocessingResult:
#         try:
#             # Implement semantic isolation logic here
#             isolated_data = self.semantic_isolation(data)
#             return PostprocessingResult(success=True, result=isolated_data)
#         except Exception as e:
#             return PostprocessingResult(success=False, error=str(e))
#
#     def semantic_isolation(self, data):
#         # Placeholder for semantic isolation logic
#         # Replace with actual implementation
#         isolation_level = self.params.get('isolation_level', 1)
#         return f"Semantic Isolation Level {isolation_level}: {data}"
#
# class ConvertToDictStep(ProcessingStep):
#     def __init__(self, params: Optional[Dict[str, Any]] = None):
#         super().__init__(params)
#         self.s2d = String2Dict()
#
#     def process(self, data: Any) -> PostprocessingResult:
#         try:
#             # Implement conversion to dict logic here using s2d.run()
#             dict_data = self.convert_to_dict(data)
#             return PostprocessingResult(success=True, result=dict_data)
#         except Exception as e:
#             return PostprocessingResult(success=False, error=str(e))
#
#     def convert_to_dict(self, data):
#         # Use String2Dict to convert string to dict
#         return self.s2d.run(data)
#
# class ExtractValueStep(ProcessingStep):
#     def process(self, data: Any) -> PostprocessingResult:
#         try:
#             # Implement value extraction logic here
#             extracted_value = self.extract_value(data)
#             return PostprocessingResult(success=True, result=extracted_value)
#         except KeyError:
#             return PostprocessingResult(success=False, error=f"Key '{self.params.get('key')}' not found.")
#         except Exception as e:
#             return PostprocessingResult(success=False, error=str(e))
#
#     def extract_value(self, data: Dict[str, Any]):
#         # Extract value based on the provided key
#         key = self.params.get('key')
#         if not key:
#             raise ValueError("No key provided for extraction.")
#         return data[key]
#
# class StringMatchValidationStep(ProcessingStep):
#     def process(self, data: Any) -> PostprocessingResult:
#         expected_string = self.params.get('expected_string', '')
#         if expected_string in str(data):
#             return PostprocessingResult(success=True, result=data)
#         else:
#             error_msg = f"Expected string '{expected_string}' not found in the result."
#             return PostprocessingResult(success=False, error=error_msg)
#
# class JsonLoadStep(ProcessingStep):
#     def process(self, data: Any) -> PostprocessingResult:
#         try:
#             json_result = json.loads(data)
#             return PostprocessingResult(success=True, result=json_result)
#         except json.JSONDecodeError as e:
#             error_msg = f"JSON loading failed: {e}"
#             return PostprocessingResult(success=False, error=error_msg)
#
# # Mapping of step types to classes
# PROCESSING_STEP_CLASSES: Dict[str, Type[ProcessingStep]] = {
#     'SemanticIsolation': SemanticIsolationStep,
#     'ConvertToDict': ConvertToDictStep,
#     'ExtractValue': ExtractValueStep,
#     'StringMatchValidation': StringMatchValidationStep,
#     'JsonLoad': JsonLoadStep,
# }
#
# class PostprocessingPipeline:
#     def __init__(self, steps_config: Optional[List[Dict[str, Any]]] = None, logger: Optional[logging.Logger] = None):
#         self.logger = logger or logging.getLogger(__name__)
#         self.steps = self._initialize_steps(steps_config or [])
#
#     def _initialize_steps(self, steps_config: List[Dict[str, Any]]) -> List[ProcessingStep]:
#         steps = []
#         for step_config in steps_config:
#             step_type = step_config.get('type')
#             params = step_config.get('params', {})
#             step_class = PROCESSING_STEP_CLASSES.get(step_type)
#             if not step_class:
#                 self.logger.error(f"Unknown processing step type: {step_type}")
#                 continue
#             step = step_class(params)
#             steps.append(step)
#         return steps
#
#     def execute(self, data: Any) -> PostprocessingResult:
#         current_data = data
#         for step in self.steps:
#             result = step.process(current_data)
#             if not result.success:
#                 # Stop processing if a step fails
#                 if self.logger:
#                     self.logger.error(f"Processing step failed: {result.error}")
#                 return result
#             current_data = result.result
#         return PostprocessingResult(success=True, result=current_data)
#
# # Postprocessor class that utilizes the pipeline
# class Postprocessor:
#     def __init__(self, logger: Optional[logging.Logger] = None, debug: bool = False):
#         self.logger = logger or logging.getLogger(__name__)
#         self.debug = debug
#
#     def postprocess(self, generation_result: GenerationResult, postprocess_config: Dict[str, Any]) -> PostprocessingResult:
#         """
#         Post-processes the LLM output based on the provided configuration.
#         Updates the generation_result in place.
#         """
#         llm_output = generation_result.content
#
#         pipeline_config = postprocess_config.get('pipeline', [])
#         pipeline = PostprocessingPipeline(steps_config=pipeline_config, logger=self.logger)
#         result = pipeline.execute(llm_output)
#
#         if result.success:
#             generation_result.content = result.result
#         else:
#             generation_result.success = False
#             generation_result.error_message = result.error
#
#         # Attach the postprocessing result to generation_result
#         generation_result.postprocessing_result = result
#
#         if self.debug:
#             if result.success:
#                 self.logger.debug(f"Postprocessing successful: {result.result}")
#             else:
#                 self.logger.error(f"Postprocessing failed: {result.error}")
#
#         return result
#
# # Example usage
# if __name__ == '__main__':
#     import logging
#     from dataclasses import dataclass, field
#     from typing import Any, Dict, Optional, Union, Literal
#
#     # Define the necessary dataclasses if not imported
#     @dataclass
#     class PostprocessingResult:
#         success: bool
#         result: Optional[Any] = None
#         error: Optional[str] = None
#
#     @dataclass
#     class GenerationResult:
#         success: bool
#         meta: Dict[str, Any] = field(default_factory=dict)
#         content: Optional[str] = None
#         raw_content: Optional[str] = None
#         elapsed_time: Optional[int] = None
#         error_message: Optional[str] = None
#         model: Optional[str] = None
#         formatted_prompt: Optional[str] = None
#         unformatted_prompt: Optional[str] = None
#         operation_name: Optional[str] = None
#         request_id: Optional[Union[str, int]] = None
#         response_type: Optional[Literal["json", "str"]] = None
#         number_of_retries: Optional[int] = None
#         postprocessing_result: Optional[PostprocessingResult] = None
#
#     # Set up basic logging
#     logging.basicConfig(level=logging.DEBUG)
#     logger = logging.getLogger('PostprocessorTest')
#
#     # Sample LLM output that is not strictly valid JSON
#     llm_output_str = "The answer is 42."
#     llm_output_dict_str = "{'answer': '42', 'explanation': 'The meaning of life.'}"  # Note single quotes
#
#     # Define pipeline configuration
#     pipeline_config = {
#         'pipeline': [
#             {
#                 'type': 'ConvertToDict',
#             },
#             {
#                 'type': 'ExtractValue',
#                 'params': {'key': 'answer'}
#             },
#             {
#                 'type': 'StringMatchValidation',
#                 'params': {'expected_string': '42'}
#             }
#         ]
#     }
#
#     # Create a sample GenerationResult
#     generation_result = GenerationResult(
#         success=True,
#         content=llm_output_dict_str,
#         meta={},
#         model='gpt-3.5-turbo',
#         formatted_prompt='...',
#         request_id=1,
#         operation_name='test_operation'
#     )
#
#     # Create a Postprocessor instance
#     postprocessor = Postprocessor(logger=logger, debug=True)
#
#     # Run postprocessing
#     postprocess_config = pipeline_config  # In this case, they are the same
#     result = postprocessor.postprocess(generation_result, postprocess_config)
#
#     # Output the result
#     if generation_result.success:
#         print("Postprocessed Content:", generation_result.content)
#     else:
#         print("Error:", generation_result.error_message)



# base_service.py

import logging
import time
import asyncio
from abc import ABC
from collections import deque
from typing import Optional, Union

from llmservice.generation_engine import GenerationEngine, GenerationRequest, GenerationResult
from llmservice.usage_stats import UsageStats


class BaseLLMService(ABC):
    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
        model_name: str = "default-model",
        yaml_file_path: Optional[str] = None,
        rpm_window_seconds: int = 60,
        max_rpm: int = 60,
        max_concurrent_requests: int = 5,
    ):
        """
        Base class for LLM services.

        :param logger: Optional logger instance.
        :param model_name: Default model name to use.
        :param yaml_file_path: Path to the YAML file containing prompts.
        :param rpm_window_seconds: Time window in seconds for RPM calculation.
        :param max_rpm: Maximum allowed Requests Per Minute.
        :param max_concurrent_requests: Maximum number of concurrent asynchronous requests.
        """
        self.logger = logger or logging.getLogger(__name__)
        self.generation_engine = GenerationEngine(logger=self.logger, model_name=model_name)
        self.usage_stats = UsageStats(model=model_name)
        self.request_id_counter = 0
        self.request_timestamps = deque()
        self.rpm_window_seconds = rpm_window_seconds
        self.max_rpm = max_rpm
        self.max_concurrent_requests = max_concurrent_requests
        self.semaphore = asyncio.Semaphore(self.max_concurrent_requests)

        if yaml_file_path:
            self.load_prompts(yaml_file_path)
        else:
            self.logger.warning("No prompts YAML file provided.")

    def _generate_request_id(self) -> int:
        """Generates a unique request ID."""
        self.request_id_counter += 1
        return self.request_id_counter

    def _store_usage(self, generation_result: GenerationResult):
        """Stores usage statistics from the generation result."""
        if generation_result and generation_result.meta:
            operation_name = generation_result.operation_name or "unknown_operation"
            # Update usage stats with operation name
            self.usage_stats.update(generation_result.meta, operation_name)
            # Log the operation name and request ID
            request_id = generation_result.request_id
            self.logger.info(
                f"Operation: {operation_name}, Request ID: {request_id}, "
                f"Input Tokens: {generation_result.meta.get('input_tokens', 0)}, "
                f"Output Tokens: {generation_result.meta.get('output_tokens', 0)}, "
                f"Total Cost: ${generation_result.meta.get('total_cost', 0):.5f}"
            )
            # Record the timestamp for RPM calculation
            self.request_timestamps.append(time.time())

            # Clean up old timestamps and log RPM
            self._clean_old_timestamps()
            rpm = self.get_current_rpm()
            self.logger.info(f"Current RPM: {rpm:.2f}")

    def _clean_old_timestamps(self):
        """Removes timestamps older than the RPM window."""
        current_time = time.time()
        while self.request_timestamps and (current_time - self.request_timestamps[0] > self.rpm_window_seconds):
            self.request_timestamps.popleft()

    def get_current_rpm(self) -> float:
        """Calculates the current Requests Per Minute (RPM)."""
        self._clean_old_timestamps()
        rpm = len(self.request_timestamps) * (60 / self.rpm_window_seconds)
        return rpm

    async def _wait_if_rate_limited(self):
        """Waits if the current RPM exceeds the maximum allowed RPM."""
        while self.get_current_rpm() >= self.max_rpm:
            wait_time = self.rpm_window_seconds - (time.time() - self.request_timestamps[0])
            wait_time = max(wait_time, 0)
            self.logger.warning(f"Rate limit reached. Waiting for {wait_time:.2f} seconds.")
            await asyncio.sleep(wait_time)
            self._clean_old_timestamps()

    def execute_generation(
        self,
        generation_request: GenerationRequest,
        operation_name: Optional[str] = None
    ) -> GenerationResult:
        """Executes the generation synchronously and stores usage statistics."""
        generation_request.operation_name = operation_name or generation_request.operation_name
        generation_request.request_id = generation_request.request_id or self._generate_request_id()

        # Rate limiting check (for synchronous calls, we can't wait asynchronously)
        if self.get_current_rpm() >= self.max_rpm:
            self.logger.error("Rate limit exceeded. Cannot proceed with synchronous execution.")
            raise Exception("Rate limit exceeded.")

        generation_result = self.generation_engine.generate_output(generation_request)
        generation_result.request_id = generation_request.request_id
        self._store_usage(generation_result)
        return generation_result

    async def execute_generation_async(
        self,
        generation_request: GenerationRequest,
        operation_name: Optional[str] = None
    ) -> GenerationResult:
        """Asynchronously executes the generation and stores usage statistics."""
        # Rate limiting check
        await self._wait_if_rate_limited()

        async with self.semaphore:
            generation_request.operation_name = operation_name or generation_request.operation_name
            generation_request.request_id = generation_request.request_id or self._generate_request_id()

            generation_result = await self.generation_engine.generate_output_async(generation_request)
            generation_result.request_id = generation_request.request_id
            self._store_usage(generation_result)
            return generation_result

    def load_prompts(self, yaml_file_path: str):
        """Loads prompts from a YAML file."""
        self.generation_engine.load_prompts(yaml_file_path)

    # Additional methods for usage stats
    def get_usage_stats(self) -> dict:
        """Returns the current usage statistics as a dictionary."""
        return self.usage_stats.to_dict()

    def reset_usage_stats(self):
        """Resets the usage statistics."""
        self.usage_stats = UsageStats(model=self.generation_engine.llm_handler.model_name)

        # Also reset request timestamps
        self.request_timestamps.clear()



# schemas.py

from dataclasses import dataclass, field, fields
from typing import Any, Dict, Optional, Union, Literal , List
import pprint

def indent_text(text, indent):
    indentation = ' ' * indent
    return '\n'.join(indentation + line for line in text.splitlines())

@dataclass
class GenerationRequest:

    data_for_placeholders: Dict[str, Any]
    unformatted_prompt: str
    model: Optional[str] = None
    output_type: Literal["json", "str"] = "str"
    use_string2dict: bool = False
    operation_name: Optional[str] = None
    postprocess_config: Optional[Dict[str, Any]] = field(default_factory=dict)
    answer_isolator_refinement_config: Optional[Dict[str, Any]] = field(default_factory=dict)
    request_id: Optional[Union[str, int]] = None
    pipeline_config: List[Dict[str, Any]] = field(default_factory=list)

@dataclass
class PipelineStepResult:
    step_type: str
    success: bool
    content_before: Any
    content_after: Any
    error_message: Optional[str] = None
    meta: Dict[str, Any] = field(default_factory=dict)




@dataclass
class PostprocessingResult:
    success: bool
    result: Optional[Any] = None
    error: Optional[str] = None
    s2d_run_status: bool = False
    s2d_run_result: Optional[Dict[str, Any]] = None
    extract_key_status: bool = False
    extract_key_result: Optional[Any] = None
    string_match_status: bool = False
    string_match_result: Optional[bool] = None
    json_load_status: bool = False
    json_load_result: Optional[Dict[str, Any]] = None
    semantic_isolation: bool = False




# class GenerationResult:
#     success: bool
#     meta: Dict[str, Any] = None  # tokens, cost, etc.
#     content: Optional[str] = None  # result
#     raw_content: Optional[str] = None  # raw result
#     elapsed_time: Optional[int] = None
#     error_message: Optional[str] = None  # rate limits
#     model: Optional[str] = None
#     formatted_prompt: Optional[str] = None  # debug
#     unformatted_prompt: Optional[str] = None  # for debug
#     operation_name: Optional[str] = None
#     request_id: Optional[Union[str, int]] = None
#     response_type: Optional[Literal["json", "str"]] = None
#     number_of_retries: Optional[int] = None  # tenacity data
#     postprocessing_result: Optional[PostprocessingResult] = None


@dataclass
class GenerationResult:
    success: bool
    meta: Dict[str, Any] = field(default_factory=dict)
    raw_content: Optional[str] = None  # Store initial LLM output
    content: Optional[Any] = None      # Final postprocessed content
    elapsed_time: Optional[float] = None
    error_message: Optional[str] = None
    model: Optional[str] = None
    formatted_prompt: Optional[str] = None
    unformatted_prompt: Optional[str] = None
    operation_name: Optional[str] = None
    request_id: Optional[Union[str, int]] = None
    response_type: Optional[str] = None
    number_of_retries: Optional[int] = None
    pipeline_steps_results: List[PipelineStepResult] = field(default_factory=list)

    def __str__(self):
        result = ["GenerationResult:"]
        for field_info in fields(self):
            field_name = field_info.name
            value = getattr(self, field_name)
            field_str = f"{field_name}:"
            if isinstance(value, (dict, list)):
                field_str += "\n" + indent_text(pprint.pformat(value, indent=4), 4)
            elif isinstance(value, str) and '\n' in value:
                # Multi-line string, indent each line
                field_str += "\n" + indent_text(value, 4)
            else:
                field_str += f" {value}"
            result.append(field_str)
        return "\n\n".join(result)

    # def __str__(self):
    #     result = ["GenerationResult:"]
    #     for field_info in fields(self):
    #         field_name = field_info.name
    #         value = getattr(self, field_name)
    #         field_str = f"{field_name}: "
    #         if isinstance(value, (dict, list)):
    #             field_str += "\n" + pprint.pformat(value, indent=4)
    #         else:
    #             field_str += str(value)
    #         result.append(field_str)
    #     return "\n".join(result)



# here is generation_engine.py

import logging

from time import time
from langchain_core.prompts import PromptTemplate
from .llm_handler import  LLMHandler
from string2dict import String2Dict
import asyncio
from indented_logger import setup_logging, log_indent
from proteas import Proteas
from .postprocessor import Postprocessor, PostprocessingResult
from .schemas import GenerationRequest, GenerationResult


setup_logging( level=logging.DEBUG,include_func=True, truncate_messages=False,min_func_name_col=100)
logger = logging.getLogger(__name__)


gpt_models_input_cost = {'gpt-4o': 5 / 1000000,
                         "gpt-4o-2024-08-06": 2.5 / 1000000,
                         'gpt-4o-mini': 0.15 / 1000000,
                         'o1-preview': 15 / 1000000,
                         'o1-mini': 3 / 1000000}

gpt_models_output_cost = {'gpt-4o': 15 / 1000000,
                          "gpt-4o-2024-08-06":   10 / 1000000,
                          'gpt-4o-mini': 0.6 / 1000000,
                          'o1-preview': 60 / 1000000,
                          'o1-mini': 12 / 1000000}









class GenerationEngine:
    def __init__(self, llm_handler=None, model_name=None, logger=None, debug=False):
        self.logger = logger or logging.getLogger(__name__)
        self.debug = debug
        self.s2d = String2Dict()

        if llm_handler:
            self.llm_handler = llm_handler
        else:
            self.llm_handler = LLMHandler(model_name=model_name)

        self.proteas = Proteas()

        self.postprocessor = Postprocessor(logger=self.logger, debug=self.debug)
        max_concurrent_requests=5
        self.semaphore = asyncio.Semaphore(max_concurrent_requests)

        if self.debug:
            pass

    def load_prompts(self, yaml_file_path):
        # yaml_file = 'prompts.yaml'
        self.proteas.load_unit_skeletons_from_yaml(yaml_file_path)

    def pick_model(self, name_of_model):
        self.llm_handler.model_name=name_of_model
        self.llm_handler.change_model( name_of_model)

    def class_logger(self):
        if self.debug:
            pass


    def craft_prompt(self, placeholder_dict, order):

        unformatted_prompt = self.proteas.craft(units=order,    placeholder_dict=placeholder_dict  )
        return unformatted_prompt


    def refine_output(self,  ):
        pass


    def answer_isolator_refiner(self, answer_to_be_refined , answer_isolater_refinement_config ):

        semantic_element_for_extraction  =answer_isolater_refinement_config["semantic_element_for_extraction"]

        order = ["answer_to_be_refined", "semantic_element_for_extraction", "answer_refiner"]
        data_for_placeholders = {"answer_to_be_refined": answer_to_be_refined,
                                 "semantic_element_for_extraction": semantic_element_for_extraction
                                 }

        unformatted_refiner_prompt = self.craft_prompt(data_for_placeholders, order)

        refiner_result = self.generate(
            unformatted_template=unformatted_refiner_prompt,
            data_for_placeholders=data_for_placeholders
        )

        if self.debug:
            self.logger.debug(f"refiner_result: {refiner_result.content}")

    #rate limit aware asyc run
    def generate_output(self, generation_request: GenerationRequest) -> GenerationResult:
        # Unpack the GenerationRequest
        placeholders = generation_request.data_for_placeholders
        unformatted_prompt = generation_request.unformatted_prompt
        postprocess_config = generation_request.postprocess_config
        answer_isolator_refinement_config = generation_request.answer_isolator_refinement_config
        operation_name = generation_request.operation_name

        generation_result = self.generate(
            unformatted_template=unformatted_prompt,
            data_for_placeholders=placeholders,
            model_name=generation_request.model
        )

        # Assign request_id and operation_name
        generation_result.request_id = generation_request.request_id
        generation_result.operation_name = generation_request.operation_name

        if generation_request.postprocess_config:
            postprocessing_result = self.postprocessor.postprocess( generation_result.content, generation_request.postprocess_config)
            generation_result.postprocessing_result = postprocessing_result
            if postprocessing_result.success:
                generation_result.content = postprocessing_result.result
            else:
                generation_result.success = False
                generation_result.error_message = postprocessing_result.error
        else:
            generation_result.postprocessing_result = None

        return generation_result

    def cost_calculator(self, input_token, output_token, model_name):
        if model_name not in gpt_models_input_cost or model_name not in gpt_models_output_cost:
            self.logger.error(f"Unsupported model name: {model_name}")
            raise ValueError(f"Unsupported model name: {model_name}")

        input_cost = gpt_models_input_cost[model_name] * int(input_token)
        output_cost = gpt_models_output_cost[model_name] * int(output_token)

        return input_cost, output_cost


    def generate(self, unformatted_template=None, data_for_placeholders=None, preprompts=None, debug=False, model_name=None):
        if preprompts:
            unformatted_prompt = self.proteas.craft(
                units=preprompts,
                placeholder_dict=data_for_placeholders,
            )

        meta = {
            "input_tokens": 0,
            "output_tokens": 0,
            "total_tokens": 0,
            "elapsed_time_for_invoke": 0,
            "input_cost": 0,
            "output_cost": 0,
            "total_cost": 0,
        }

        t0 = time()

        from langchain_core.prompts.string import get_template_variables

        existing_placeholders = get_template_variables(unformatted_template, "f-string")
        missing_placeholders = set(existing_placeholders) - set(data_for_placeholders.keys())

        if missing_placeholders:
            raise ValueError(f"Missing data for placeholders: {missing_placeholders}")

        filtered_data = {key: value for key, value in data_for_placeholders.items() if key in existing_placeholders}

        prompt = PromptTemplate.from_template(unformatted_template)
        formatted_prompt = prompt.format(**filtered_data)

        t1 = time()

        # Initialize LLMHandler with the model_name
        llm_handler = LLMHandler(model_name=model_name or self.llm_handler.model_name)

        r, success = llm_handler.invoke(prompt=formatted_prompt)

        if not success:
            return GenerationResult(success=success,
                                    meta=meta,
                                    content=None,
                                    elapsed_time=0,
                                    error_message="LLM invocation failed",
                                    model=llm_handler.model_name,
                                    formatted_prompt=formatted_prompt)

        t2 = time()
        elapsed_time_for_invoke = t2 - t1
        meta["elapsed_time_for_invoke"] = elapsed_time_for_invoke

        if llm_handler.OPENAI_MODEL:
            try:
                meta["input_tokens"] = r.usage_metadata["input_tokens"]
                meta["output_tokens"] = r.usage_metadata["output_tokens"]
                meta["total_tokens"] = r.usage_metadata["total_tokens"]
            except KeyError as e:
                return "error", formatted_prompt, meta, False

            input_cost, output_cost = self.cost_calculator(meta["input_tokens"], meta["output_tokens"],
                                                           llm_handler.model_name)
            meta["input_cost"] = input_cost
            meta["output_cost"] = output_cost
            meta["total_cost"] = input_cost + output_cost

        return GenerationResult(success=success,
                                meta=meta,
                                content=r.content,
                                elapsed_time=elapsed_time_for_invoke,
                                error_message=None,
                                model=llm_handler.model_name,
                                formatted_prompt=formatted_prompt)

    async def generate_async(self, unformatted_template=None, data_for_placeholders=None, preprompts=None, debug=False,
                             model_name=None):

        async with self.semaphore:
            # Similar to generate, but uses async methods
            if preprompts:
                unformatted_prompt = self.proteas.craft(
                    units=preprompts,
                    placeholder_dict=data_for_placeholders,
                )

            meta = {
                "input_tokens": 0,
                "output_tokens": 0,
                "total_tokens": 0,
                "elapsed_time_for_invoke": 0,
                "input_cost": 0,
                "output_cost": 0,
                "total_cost": 0,
            }

            t0 = time()

            from langchain_core.prompts.string import get_template_variables

            existing_placeholders = get_template_variables(unformatted_template, "f-string")
            missing_placeholders = set(existing_placeholders) - set(data_for_placeholders.keys())

            if missing_placeholders:
                raise ValueError(f"Missing data for placeholders: {missing_placeholders}")

            filtered_data = {key: value for key, value in data_for_placeholders.items() if key in existing_placeholders}

            prompt = PromptTemplate.from_template(unformatted_template)
            formatted_prompt = prompt.format(**filtered_data)

            t1 = time()

            # Initialize LLMHandler with the model_name
            llm_handler = LLMHandler(model_name=model_name or self.llm_handler.model_name)

            # Use async invoke
            r, success = await llm_handler.invoke_async(prompt=formatted_prompt)

            if not success:
                return GenerationResult(success=success,
                                        meta=meta,
                                        content=None,
                                        elapsed_time=0,
                                        error_message="LLM invocation failed",
                                        model=llm_handler.model_name,
                                        formatted_prompt=formatted_prompt)

            t2 = time()
            elapsed_time_for_invoke = t2 - t1
            meta["elapsed_time_for_invoke"] = elapsed_time_for_invoke

            if llm_handler.OPENAI_MODEL:
                try:
                    meta["input_tokens"] = r.usage_metadata["input_tokens"]
                    meta["output_tokens"] = r.usage_metadata["output_tokens"]
                    meta["total_tokens"] = r.usage_metadata["total_tokens"]
                except KeyError as e:
                    return "error", formatted_prompt, meta, False

                input_cost, output_cost = self.cost_calculator(meta["input_tokens"], meta["output_tokens"],
                                                               llm_handler.model_name)
                meta["input_cost"] = input_cost
                meta["output_cost"] = output_cost
                meta["total_cost"] = input_cost + output_cost

            return GenerationResult(success=success,
                                    meta=meta,
                                    content=r.content,
                                    elapsed_time=elapsed_time_for_invoke,
                                    error_message=None,
                                    model=llm_handler.model_name,
                                    formatted_prompt=formatted_prompt)



    async def generate_output_async(self, generation_request: GenerationRequest) -> GenerationResult:
        # Unpack the GenerationRequest
        placeholders = generation_request.data_for_placeholders
        unformatted_prompt = generation_request.unformatted_prompt
        operation_name = generation_request.operation_name

        # Use async generate method
        generation_result = await self.generate_async(
            unformatted_template=unformatted_prompt,
            data_for_placeholders=placeholders,
            model_name=generation_request.model  # Pass model_name
        )

        # Assign request_id and operation_name
        generation_result.request_id = generation_request.request_id
        generation_result.operation_name = generation_request.operation_name

        if generation_request.postprocess_config:
            postprocessing_result = self.postprocessor.postprocess(
                generation_result.content, generation_request.postprocess_config)
            generation_result.postprocessing_result = postprocessing_result
            if postprocessing_result.success:
                generation_result.content = postprocessing_result.result
            else:
                generation_result.success = False
                generation_result.error_message = postprocessing_result.error
        else:
            generation_result.postprocessing_result = None

        return generation_result



    async def invoke_llm_async(self, formatted_prompt):
        """Asynchronously invokes the LLM with the formatted prompt."""
        t_start = time()
        response, success = await self.llm_handler.invoke_async(prompt=formatted_prompt)
        t_end = time()
        elapsed_time = t_end - t_start

        if not success:
            self.logger.error(f"LLM invocation failed with response: {response}")
            return None, success, elapsed_time

        return response, success, elapsed_time


def main():
    import logging

    # Set up basic logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger('GenerationEngineExample')

    # Initialize the GenerationEngine with a specific model (e.g., 'gpt-4')
    gen_engine = GenerationEngine(logger=logger, model_name='gpt-4o')

    # Define the unformatted prompt directly
    unformatted_prompt = "Translate the following text to French:\n\n{text_to_translate}"

    # Data for placeholders
    data_for_placeholders = {
        'text_to_translate': "Hello, how are you?"
    }

    # Create a GenerationRequest
    generation_request = GenerationRequest(
        data_for_placeholders=data_for_placeholders,
        unformatted_prompt=unformatted_prompt,
        operation_name='translate_to_french',
        request_id='example_request_1'
    )

    # Generate the output
    generation_result = gen_engine.generate_output(generation_request)

    # Check if generation was successful
    if generation_result.success:
        print("Generated content:")
        print(generation_result.content)
    else:
        print("Generation failed:")
        print(f"Error: {generation_result.error_message}")

if __name__ == '__main__':
    main()



# generation_engine.py

import logging
import time
from typing import Optional, Dict, Any, List, Union
from dataclasses import dataclass, field

from llmservice.llm_handler import LLMHandler  # Ensure this is correctly imported
from string2dict import String2Dict  # Ensure this is installed and available
from proteas import Proteas  # Ensure this is installed and available
from langchain_core.prompts import PromptTemplate
from langchain_core.prompts.string import get_template_variables

from .schemas import GenerationRequest, GenerationResult, PostprocessingResult , PipelineStepResult


logger = logging.getLogger(__name__)

# Costs per model (example values, adjust as needed)
gpt_models_input_cost = {'gpt-4o': 5 / 1000000,
                         "gpt-4o-2024-08-06": 2.5 / 1000000,
                         'gpt-4o-mini': 0.15 / 1000000,
                         'o1-preview': 15 / 1000000,
                         'o1-mini': 3 / 1000000}

gpt_models_output_cost = {'gpt-4o': 15 / 1000000,
                          "gpt-4o-2024-08-06":   10 / 1000000,
                          'gpt-4o-mini': 0.6 / 1000000,
                          'o1-preview': 60 / 1000000,
                          'o1-mini': 12 / 1000000}


class GenerationEngine:
    def __init__(self, llm_handler=None, model_name=None, debug=False):
        self.logger = logging.getLogger(__name__)
        self.debug = debug
        self.s2d = String2Dict()

        if llm_handler:
            self.llm_handler = llm_handler
        else:
            self.llm_handler = LLMHandler(model_name=model_name, logger=self.logger)

        self.proteas = Proteas()

        if self.debug:
            self.logger.setLevel(logging.DEBUG)

        # Define the semantic isolation prompt template
        self.semantic_isolation_prompt_template = """
Here is the text answer which includes the main desired information as well as some additional information: {answer_to_be_refined}
Here is the semantic element which should be used for extraction: {semantic_element_for_extraction}

From the given text answer, isolate and extract the semantic element.
Provide the answer strictly in the following JSON format, do not combine anything, remove all introductory or explanatory text that is not part of the semantic element:

'answer': 'here_is_isolated_answer'
"""

    def _debug(self, message):

        if self.debug:
            self.logger.debug(message)

    def load_prompts(self, yaml_file_path):
        """Loads prompts from a YAML file using Proteas."""
        self.proteas.load_unit_skeletons_from_yaml(yaml_file_path)

    def craft_prompt(self, placeholder_dict: Dict[str, Any], order: Optional[list] = None) -> str:
        """
        Crafts the prompt using Proteas with the given placeholders and order.

        :param placeholder_dict: Dictionary of placeholder values.
        :param order: Optional list specifying the order of units.
        :return: Unformatted prompt string.
        """
        unformatted_prompt = self.proteas.craft(units=order, placeholder_dict=placeholder_dict)
        return unformatted_prompt

    def cost_calculator(self, input_token, output_token, model_name):
        if model_name not in gpt_models_input_cost or model_name not in gpt_models_output_cost:
            self.logger.error(f"Unsupported model name: {model_name}")
            raise ValueError(f"Unsupported model name: {model_name}")

        input_cost = gpt_models_input_cost[model_name] * int(input_token)
        output_cost = gpt_models_output_cost[model_name] * int(output_token)

        return input_cost, output_cost

    def generate_output(self, generation_request: GenerationRequest) -> GenerationResult:
        """
        Synchronously generates the output and processes postprocessing.

        :param generation_request: GenerationRequest object containing all necessary data.
        :return: GenerationResult object with the output and metadata.
        """
        # Unpack the GenerationRequest
        placeholders = generation_request.data_for_placeholders
        unformatted_prompt = generation_request.unformatted_prompt

        # Generate the output synchronously
        generation_result = self.generate(
            unformatted_template=unformatted_prompt,
            data_for_placeholders=placeholders,
            model_name=generation_request.model,
            request_id=generation_request.request_id,
            operation_name=generation_request.operation_name
        )

        if not generation_result.success:
            return generation_result

        # Process the output using the pipeline
        if generation_request.pipeline_config:
            generation_result = self.execute_pipeline(generation_result, generation_request.pipeline_config)
        else:
            # No postprocessing; assign raw_content to content
            generation_result.content = generation_result.raw_content

        return generation_result

    def execute_pipeline(self, generation_result: GenerationResult, pipeline_config: List[Dict[str, Any]]) -> GenerationResult:
        """
        Executes the processing pipeline on the generation result.

        :param generation_result: The initial GenerationResult from the LLM.
        :param pipeline_config: List of processing steps.
        :return: Updated GenerationResult after processing.
        """
        current_content = generation_result.raw_content
        for step_config in pipeline_config:
            step_type = step_config.get('type')
            params = step_config.get('params', {})
            method_name = f"process_{step_type.lower()}"
            processing_method = getattr(self, method_name, None)
            step_result = PipelineStepResult(
                step_type=step_type,
                success=False,
                content_before=current_content,
                content_after=None
            )
            if processing_method:
                try:
                    content_after = processing_method(current_content, **params)
                    step_result.success = True
                    step_result.content_after = content_after
                    current_content = content_after  # Update current_content for next step
                except Exception as e:
                    step_result.success = False
                    step_result.error_message = str(e)
                    generation_result.success = False
                    generation_result.error_message = f"Processing step '{step_type}' failed: {e}"
                    self.logger.error(generation_result.error_message)
                    # Record the failed step and exit the pipeline
                    generation_result.pipeline_steps_results.append(step_result)
                    return generation_result
            else:
                step_result.success = False
                error_msg = f"Unknown processing step type: {step_type}"
                step_result.error_message = error_msg
                generation_result.success = False
                generation_result.error_message = error_msg
                self.logger.error(generation_result.error_message)
                # Record the failed step and exit the pipeline
                generation_result.pipeline_steps_results.append(step_result)
                return generation_result
            # Record the successful step
            generation_result.pipeline_steps_results.append(step_result)

        # Update the final content
        generation_result.content = current_content
        return generation_result

    # Define processing methods
    def process_semanticisolation(self, content: str, semantic_element_for_extraction: str) -> str:
        """
        Processes content using semantic isolation.

        :param content: The content to process.
        :param semantic_element_for_extraction: The semantic element to extract.
        :return: The isolated semantic element.
        """
        answer_to_be_refined = content

        data_for_placeholders = {
            "answer_to_be_refined": answer_to_be_refined,
            "semantic_element_for_extraction": semantic_element_for_extraction,
        }
        unformatted_refiner_prompt = self.semantic_isolation_prompt_template

        refiner_result = self.generate(
            unformatted_template=unformatted_refiner_prompt,
            data_for_placeholders=data_for_placeholders
        )

        if not refiner_result.success:
            raise ValueError(f"Semantic isolation failed: {refiner_result.error_message}")

        # Parse the LLM response to extract 'answer'
        s2d_result = self.s2d.run(refiner_result.raw_content)
        isolated_answer = s2d_result.get('answer')
        if isolated_answer is None:
            raise ValueError("Isolated answer not found in the LLM response.")

        return isolated_answer

    def process_converttodict(self, content: Any) -> Dict[str, Any]:
        """
        Converts content to a dictionary.

        :param content: The content to convert.
        :return: The content as a dictionary.
        """
        if isinstance(content, dict):
            return content  # Already a dict
        return self.s2d.run(content)

    def process_extractvalue(self, content: Dict[str, Any], key: str) -> Any:
        """
        Extracts a value from a dictionary.

        :param content: The dictionary content.
        :param key: The key to extract.
        :return: The extracted value.
        """
        if key not in content:
            raise KeyError(f"Key '{key}' not found in content.")
        return content[key]

    def process_stringmatchvalidation(self, content: str, expected_string: str) -> str:
        """
        Validates that the expected string is present in the content.

        :param content: The content to validate.
        :param expected_string: The expected string to find.
        :return: The original content if validation passes.
        """
        if expected_string not in content:
            raise ValueError(f"Expected string '{expected_string}' not found in content.")
        return content

    def process_jsonload(self, content: str) -> Dict[str, Any]:
        """
        Loads content as JSON.

        :param content: The content to load.
        :return: The content as a JSON object.
        """
        import json
        try:
            return json.loads(content)
        except json.JSONDecodeError as e:
            raise ValueError(f"JSON loading failed: {e}")

    # Implement the generate method
    def generate(self, unformatted_template, data_for_placeholders, model_name=None, request_id=None, operation_name=None) -> GenerationResult:
        """
        Generates content using the LLMHandler.

        :param unformatted_template: The unformatted prompt template.
        :param data_for_placeholders: Data to fill the placeholders.
        :param model_name: Model name to use.
        :param request_id: Optional request ID.
        :param operation_name: Optional operation name.
        :return: GenerationResult object.
        """
        meta = {
            "input_tokens": 0,
            "output_tokens": 0,
            "total_tokens": 0,
            "elapsed_time_for_invoke": 0,
            "input_cost": 0,
            "output_cost": 0,
            "total_cost": 0,
        }

        t0 = time.time()

        # Validate placeholders
        existing_placeholders = get_template_variables(unformatted_template, "f-string")
        missing_placeholders = set(existing_placeholders) - set(data_for_placeholders.keys())

        if missing_placeholders:
            raise ValueError(f"Missing data for placeholders: {missing_placeholders}")

        # Format the prompt
        prompt_template = PromptTemplate.from_template(unformatted_template)
        formatted_prompt = prompt_template.format(**data_for_placeholders)

        t1 = time.time()

        # Initialize LLMHandler with the model_name
        llm_handler = LLMHandler(model_name=model_name or self.llm_handler.model_name, logger=self.logger)

        # Invoke the LLM synchronously
        r, success = llm_handler.invoke(prompt=formatted_prompt)

        if not success:
            return GenerationResult(
                success=False,
                meta=meta,
                raw_content=None,
                content=None,
                elapsed_time=0,
                error_message="LLM invocation failed",
                model=llm_handler.model_name,
                formatted_prompt=formatted_prompt,
                request_id=request_id,
                operation_name=operation_name
            )

        t2 = time.time()
        elapsed_time_for_invoke = t2 - t1
        meta["elapsed_time_for_invoke"] = elapsed_time_for_invoke

        if llm_handler.OPENAI_MODEL:
            try:
                meta["input_tokens"] = r.usage_metadata["input_tokens"]
                meta["output_tokens"] = r.usage_metadata["output_tokens"]
                meta["total_tokens"] = r.usage_metadata["total_tokens"]
            except KeyError as e:
                return GenerationResult(
                    success=False,
                    meta=meta,
                    raw_content=None,
                    content=None,
                    elapsed_time=elapsed_time_for_invoke,
                    error_message="Token usage metadata missing",
                    model=llm_handler.model_name,
                    formatted_prompt=formatted_prompt,
                    unformatted_prompt=unformatted_template,
                    request_id=request_id,
                    operation_name=operation_name
                )

            input_cost, output_cost = self.cost_calculator(
                meta["input_tokens"], meta["output_tokens"], llm_handler.model_name)
            meta["input_cost"] = input_cost
            meta["output_cost"] = output_cost
            meta["total_cost"] = input_cost + output_cost

        return GenerationResult(
            success=True,
            meta=meta,
            raw_content=r.content,  # Assign initial LLM output
            content=None,           # Will be assigned after postprocessing
            elapsed_time=elapsed_time_for_invoke,
            error_message=None,
            model=llm_handler.model_name,
            formatted_prompt=formatted_prompt,
            unformatted_prompt=unformatted_template,
            request_id=request_id,
            operation_name=operation_name
        )

# Main function for testing
def main():
    import logging

    import sys

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        stream=sys.stdout
    )


    generation_engine = GenerationEngine(model_name='gpt-4o')

    placeholders = {'input_text': 'Patient shows symptoms of severe headache and nausea.'}
    unformatted_prompt = 'Provide a summary of the following clinical note: {input_text}'

    pipeline_config = [
        {
            'type': 'SemanticIsolation',
            'params': {
                'semantic_element_for_extraction': 'symptoms'
            }
        },
        # You can add more steps here if needed
    ]

    gen_request = GenerationRequest(
        data_for_placeholders=placeholders,
        unformatted_prompt=unformatted_prompt,
        model='gpt-4o',
        pipeline_config=pipeline_config,
        request_id=3,
        operation_name='extract_symptoms'
    )

    generation_result = generation_engine.generate_output(gen_request)

    if generation_result.success:
        logger.info("Final Result:")
        logger.info(generation_result.content)
        logger.info("Raw LLM Output:")
        logger.info(generation_result.raw_content)
    else:
        logger.info("Error:")
        logger.info(generation_result.error_message)

    logger.info("Pipeline Steps Results")
    for step_result in generation_result.pipeline_steps_results:
        logger.info(f"Step {step_result.step_type}")
        logger.info(f"Success {step_result.success}")

        if step_result.success:
            logger.info(f"Content Before: {step_result.content_before}")
            logger.info(f"Content After: {step_result.content_after}")
        else:
            logger.info(f"Error: {step_result.error_message}")

if __name__ == '__main__':
    main()
